import torch
from torch import nn
import torch.nn.functional as F

from ._ETP import ETP
from ._model_utils import pairwise_euclidean_distance


class fastopic(nn.Module):
    def __init__(self,
                 num_topics: int,
                 theta_temp: float=1.0,
                 DT_alpha: float=3.0,
                 TW_alpha: float=2.0
                ):
        super().__init__()

        self.num_topics = num_topics
        self.DT_alpha = DT_alpha
        self.TW_alpha = TW_alpha
        self.theta_temp = theta_temp

        self.epsilon = 1e-12

    def init(self,
             vocab_size: int,
             embed_size: int,
             _fitted: bool = False,
             pre_vocab: list=None,
             vocab: list=None
            ):

        if _fitted:
            topic_embeddings = self.topic_embeddings.data
            assert topic_embeddings.shape == (self.num_topics, embed_size)
            topic_weights = self.topic_weights.data
            del self.topic_weights
        else:
            topic_embeddings = F.normalize(nn.init.trunc_normal_(torch.empty((self.num_topics, embed_size))))
            topic_weights = (torch.ones(self.num_topics) / self.num_topics).unsqueeze(1)

        self.topic_embeddings = nn.Parameter(topic_embeddings)
        self.topic_weights = nn.Parameter(topic_weights)

        word_embeddings = F.normalize(nn.init.trunc_normal_(torch.empty(vocab_size, embed_size)))
        if _fitted:
            pre_word_embeddings = self.word_embeddings.data
            word_weights = torch.zeros(vocab_size, 1)
            pre_norm_word_weights = F.softmax(self.word_weights.data, dim=0)
            del self.word_embeddings
            del self.word_weights

            for i, word in enumerate(vocab):
                if word in pre_vocab:
                    pre_word_idx = pre_vocab.index(word)
                    word_embeddings[i] = pre_word_embeddings[pre_word_idx]
                    word_weights[i] = pre_norm_word_weights[pre_word_idx]

            left_avg = (1.0 - word_weights.sum()) / word_weights.nonzero().size(0)
            word_weights[word_weights == 0] = left_avg

            word_weights = torch.log(word_weights)
            word_weights = word_weights - word_weights.mean()

        else:
            word_weights = (torch.ones(vocab_size) / vocab_size).unsqueeze(1)

        self.word_embeddings = nn.Parameter(word_embeddings)
        self.word_weights = nn.Parameter(word_weights)

        self.DT_ETP = ETP(self.DT_alpha, init_b_dist=self.topic_weights)
        self.TW_ETP = ETP(self.TW_alpha, init_b_dist=self.word_weights)

    def get_transp_DT(self, doc_embeddings):

        topic_embeddings = self.topic_embeddings.detach().to(doc_embeddings.device)
        _, transp = self.DT_ETP(doc_embeddings, topic_embeddings)

        return transp.detach().cpu().numpy()

    # only for testing
    def get_beta(self):
        _, transp_TW = self.TW_ETP(self.topic_embeddings, self.word_embeddings)
        # use transport plan as beta
        beta = transp_TW * transp_TW.shape[0]

        return beta

    # only for testing
    def get_theta(self,
            doc_embeddings,
            train_doc_embeddings
        ):
        topic_embeddings = self.topic_embeddings.detach().to(doc_embeddings.device)
        dist = pairwise_euclidean_distance(doc_embeddings, topic_embeddings)
        train_dist = pairwise_euclidean_distance(train_doc_embeddings, topic_embeddings)

        exp_dist = torch.exp(-dist / self.theta_temp)
        exp_train_dist = torch.exp(-train_dist / self.theta_temp)

        theta = exp_dist / (exp_train_dist.sum(0))
        theta = theta / theta.sum(1, keepdim=True)

        return theta

    def forward(self, train_bow, doc_embeddings):
        loss_DT, transp_DT = self.DT_ETP(doc_embeddings, self.topic_embeddings)
        loss_TW, transp_TW = self.TW_ETP(self.topic_embeddings, self.word_embeddings)

        loss_ETP = loss_DT + loss_TW

        theta = transp_DT * transp_DT.shape[0]
        beta = transp_TW * transp_TW.shape[0]

        # Dual Semantic-relation Reconstruction (DSR)
        recon = torch.matmul(theta, beta)
        loss_DSR = -(train_bow * (recon + self.epsilon).log()).sum(axis=1).mean()

        loss = loss_DSR + loss_ETP

        rst_dict = {
            'loss': loss,
        }

        return rst_dict



import pytest
import shutil
from numpy.testing import assert_almost_equal
from sklearn.datasets import fetch_20newsgroups

import sys
sys.path.append("../")

from fastopic import FASTopic
from topmost import Preprocess

@pytest.fixture
def cache_path():
    return "./pytest_cache/"


@pytest.fixture
def num_topics():
    return 10


def test(cache_path, num_topics):
    docs = fetch_20newsgroups(subset='all', remove=('headers', 'footers', 'quotes'))['data']
    docs = docs[:100]

    model = FASTopic(num_topics)
    top_words, doc_topic_dist = model.fit_transform(docs, epochs=1)

    preprocess = Preprocess(vocab_size=10000, stopwords="English")

    model = FASTopic(num_topics, preprocess)
    top_words, doc_topic_dist = model.fit_transform(docs, epochs=1)
    beta = model.get_beta()

    path = f"{cache_path}/tmp_save/fastopic.zip"
    model.save(path)

    loaded_model = FASTopic.from_pretrained(path)

    loaded_beta = loaded_model.get_beta()
    assert_almost_equal(beta, loaded_beta)

    loaded_doc_topic_dist = loaded_model.transform(docs)
    assert_almost_equal(doc_topic_dist, loaded_doc_topic_dist)

    # Keep training
    loaded_model.fit_transform(docs, epochs=1)

    shutil.rmtree(f"{cache_path}/tmp_save")



import numpy as np
import pytest

import sys
sys.path.append('../')

from fastopic import FASTopic
from topmost import download_dataset, DynamicDataset


@pytest.fixture
def cache_path():
    return './pytest_cache/'


@pytest.fixture
def num_topics():
    return 10


def model_test(model, dataset, num_topics):
    docs = dataset.train_texts

    top_words, train_theta = model.fit_transform(docs, epochs=1)
    top_words, train_theta = model.fit_transform(docs, epochs=1)

    test_theta = model.transform(dataset.test_texts)

    assert len(top_words) == num_topics
    assert train_theta.shape[0] == len(docs)
    assert test_theta.shape[0] == len(dataset.test_texts)

    model.get_topic(0)
    model.visualize_topic()
    model.visualize_topic_hierarchy()
    model.visualize_topic_weights(top_n=10, height=500)

    time_slices = dataset.train_times
    act = model.topic_activity_over_time(time_slices)
    model.visualize_topic_activity(top_n=10, topic_activity=act, time_slices=time_slices)

    assert act.shape == (num_topics, len(np.unique(time_slices)))

def test_models(cache_path, num_topics):
    download_dataset("NYT", cache_path=f"{cache_path}/datasets")
    dataset = DynamicDataset(f"{cache_path}/datasets/NYT", as_tensor=False)

    model = FASTopic(num_topics, verbose=True)
    model_test(model, dataset, num_topics)

    model = FASTopic(
        num_topics=num_topics,
        verbose=True,
        low_memory=True,
        low_memory_batch_size=len(dataset.train_texts) // 2,
    )
    model_test(model, dataset, num_topics)



import numpy as np
import itertools

from sklearn.metrics.pairwise import cosine_similarity, euclidean_distances
from scipy.cluster import hierarchy as sch

import plotly.figure_factory as ff
import plotly.graph_objects as go
from plotly.subplots import make_subplots

from typing import Callable, List, Union


def wrap_topic_idx(
        topic_model,
        top_n: int=None,
        topic_idx: List[int]=None
    ):

    topic_weights = topic_model.get_topic_weights()

    if top_n is None and topic_idx is None:
        top_n = 5
        topic_idx = np.argsort(topic_weights)[:-(top_n + 1):-1]
    elif top_n is not None:
        assert (top_n > 0) and (topic_idx is None)
        topic_idx = np.argsort(topic_weights)[:-(top_n + 1):-1]

    return topic_idx


def visualize_topic(topic_model,
                    top_n: int=None,
                    topic_idx: List[int]=None,
                    n_label_words=5,
                    width: int = 250,
                    height: int = 250,
                    topic_labels: List[str]=None,
                ):

    topic_idx = wrap_topic_idx(topic_model, top_n, topic_idx)

    top_words = topic_model.top_words
    beta = topic_model.get_beta()

    if topic_labels is None:
        subplot_titles = [f"Topic {i}" for i in topic_idx]
    else:
        assert len(topic_labels) == len(topic_idx), "Labels length must match the number of topics."
        subplot_titles = [topic_labels[i] for i in topic_idx]

    columns = 4
    rows = int(np.ceil(len(topic_idx) / columns))

    colors = itertools.cycle(["#D55E00", "#0072B2", "#CC79A7", "#E69F00", "#56B4E9", "#009E73", "#F0E442"])

    fig = make_subplots(rows=rows,
                        cols=columns,
                        shared_xaxes=False,
                        horizontal_spacing=.1,
                        subplot_titles=subplot_titles,
                    )

    row = 1
    column = 1
    for i in topic_idx:
        words = top_words[i].split()[:n_label_words][::-1]
        scores = np.sort(beta[i])[:-(n_label_words + 1):-1][::-1]

        fig.add_trace(
                go.Bar(x=scores,
                    y=words,
                    orientation='h',
                    marker_color=next(colors)
                ),
                row=row,
                col=column
            )

        if column == columns:
            column = 1
            row += 1
        else:
            column += 1

    fig.update_layout(
        template="plotly_white",
        showlegend=False,
        title={
            'text': f"Topic-Word Distributions",
            'x': .5,
            'xanchor': 'center',
            'yanchor': 'top',
            'font': dict(
                size=22,
                color="Black")
        },
        width=width * 4,
        height=height * rows if rows > 1 else height * 1.3,
        hoverlabel=dict(
            bgcolor="white",
            font_size=16,
            font_family="Rockwell"
        ),
    )

    fig.update_xaxes(showgrid=True)
    fig.update_yaxes(showgrid=True)

    return fig


def visualize_activity(topic_model,
                       topic_activity: np.ndarray,
                       time_slices: Union[np.ndarray, List],
                       top_n: int=None,
                       topic_idx: List[int]=None,
                       n_label_words:int=5,
                       title: str="<b>Topics Activity over Time</b>",
                       width: int=1000,
                       height: int=600,
                       topic_labels: List[str]=None,
                    ):

    topic_idx = wrap_topic_idx(topic_model, top_n, topic_idx)

    colors = ["#E69F00", "#56B4E9", "#009E73", "#F0E442", "#D55E00", "#0072B2", "#CC79A7"]

    fig = go.Figure()
    topic_top_words = topic_model.top_words

    if topic_labels is None:
        topic_labels = []
        for i, words in enumerate(topic_top_words):
            topic_labels.append(f"{i}_{'_'.join(words.split()[:n_label_words])}")
    else:
        assert len(topic_labels) == topic_activity.shape[0], "Labels length must match the number of topics."

    labels = np.unique(time_slices).tolist()

    for i, k in enumerate(topic_idx):

        fig.add_trace(go.Scatter(
            x=labels,
            y=topic_activity[k].tolist(),
            mode='lines',
            marker_color=colors[i % 7],
            hoverinfo="text",
            name=topic_labels[k],
            hovertext=topic_labels[k])
        )

    # Styling of the visualization
    fig.update_xaxes(showgrid=True)
    fig.update_yaxes(showgrid=True)
    fig.update_layout(
        yaxis_title="Topic Weight",
        title={
            'text': f"{title}",
            'y': .95,
            'x': 0.40,
            'xanchor': 'center',
            'yanchor': 'top',
            'font': dict(
                size=22,
                color="Black")
        },
        template="simple_white",
        width=width,
        height=height,
        hoverlabel=dict(
            bgcolor="white",
            font_size=16,
            font_family="Rockwell"
        )
    )

    return fig


def visualize_topic_weights(topic_model,
                            top_n: int=50,
                            topic_idx: List[int]=None,
                            n_label_words: int=5,
                            title: str="<b>Topic Weights</b>",
                            width: int=1000,
                            height: int=1000,
                            _sort: bool=True,
                            topic_labels: List[str]=None,
                        ):

    topic_weights = topic_model.get_topic_weights()
    topic_idx = wrap_topic_idx(topic_model, top_n, topic_idx)

    labels = []
    vals = []
    topic_top_words = topic_model.top_words
    
    for i in topic_idx:
        words = topic_top_words[i]
        labels.append(topic_labels[i] if topic_labels is not None else f"{i}_{'_'.join(words.split()[:n_label_words])}")
        vals.append(topic_weights[i])

    if _sort:
        sorted_idx = np.argsort(vals)
        labels = np.asarray(labels)[sorted_idx].tolist()
        vals = np.asarray(vals)[sorted_idx].tolist()

    # Create Figure
    fig = go.Figure(go.Bar(
        x=vals,
        y=labels,
        marker=dict(
            color='#C8D2D7',
            line=dict(
                color='#6E8484',
                width=1),
        ),
        orientation='h')
    )

    fig.update_layout(
        xaxis_title="Weight",
        title={
            'text': f"{title}",
            'y': .95,
            'x': 0.5,
            'xanchor': 'center',
            'yanchor': 'top',
            'font': dict(
                size=22,
                color="Black")
        },
        template="simple_white",
        width=width,
        height=height,
        hoverlabel=dict(
            bgcolor="white",
            font_size=16,
            font_family="Rockwell"
        ),
    )

    return fig


def visualize_hierarchy(topic_model,
                        orientation: str = "left",
                        width: int = 1000,
                        height: int = 1000,
                        linkage_function: Callable = None,
                        distance_function: Callable = None,
                        n_label_words: int = 5,
                        color_threshold: int = None,
                        topic_labels: List[str] = None, # labels for the topics
                    ):

    topic_embeddings = topic_model.topic_embeddings

    if distance_function is None:
        # distance_function = lambda x: 1 - cosine_similarity(x)
        distance_function = euclidean_distances

    if linkage_function is None:
        linkage_function = lambda x: sch.linkage(x, 'ward', optimal_ordering=True)

    topic_top_words = topic_model.top_words
    
    if topic_labels is None:
        topic_labels = []
        for i, words in enumerate(topic_top_words):
            topic_labels.append(f"{i}_{'_'.join(words.split()[:n_label_words])}")
    else:
        assert len(topic_labels) == topic_embeddings.shape[0], "Labels length must match the number of topics."
        
    fig = ff.create_dendrogram(
        topic_embeddings,
        orientation=orientation,
        labels=topic_labels,
        distfun=distance_function,
        linkagefun=linkage_function,
        color_threshold=color_threshold
    )

    fig.update_layout({'width': width, 'height': height})

    return fig



import numpy as np
import torch
import pandas as pd
from collections import defaultdict
from pathlib import Path
from tqdm import tqdm


from . import _plot
from ._fastopic import fastopic
from ._utils import Logger, assert_fitted, check_fitted, DocEmbedModel, Dataset, get_top_words

from typing import List, Union, Callable


logger = Logger("WARNING")


class FASTopic:
    def __init__(
        self,
        num_topics: int,
        preprocess: Callable = None,
        num_top_words: int = 15,
        device: str = None,
        normalize_embeddings: bool = False,
        doc_embed_model: Union[str, callable] = "all-MiniLM-L6-v2",
        DT_alpha: float = 3.0,
        TW_alpha: float = 2.0,
        theta_temp: float = 1.0,
        low_memory: bool = False,
        low_memory_batch_size: int = None,
        verbose: bool = False,
        log_interval: int = 10,
    ):
        """FASTopic initialization.

        Args:
            num_topics: The number of topics.
            preprocess: preprocess class from topmost.preprocess.Preprocess or user-defined module.
            doc_embed_model: The used document embedding model.
                             This can be your callable model that implements `.encode(docs)`.
                             This can also be a model name in sentence-transformers.
                             The default is "all-MiniLM-L6-v2".
            num_top_words: The number of top words to be returned in topics.
            DT_alpha: The sinkhorn alpha between document embeddings and topic embeddings.
            TW_alpha: The sinkhorn alpha between topic embeddings and word embeddings.
            theta_temp: The temperature parameter of the softmax used
                        to compute doc topic distributions during testing.
            device: The device.
            normalize_embeddings: Set this to True to normalize document embeddings.
                                This parameter may not be effective
                                when you use your own document embedding model.
            log_interval: The interval to print logs during training.
            verbose: Changes the verbosity of the model, Set to True if you want
                     to track the stages of the model.
        """

        if device is None:
            device = 'cuda' if torch.cuda.is_available() else 'cpu'
        self.device = device

        self.num_top_words = num_top_words
        self.doc_embed_model = doc_embed_model
        self.normalize_embeddings = normalize_embeddings
        self.low_memory = low_memory
        self.low_memory_batch_size = low_memory_batch_size

        self.beta = None
        self.train_theta = None
        self.model = fastopic(num_topics, theta_temp, DT_alpha, TW_alpha)

        if preprocess is None:
            from topmost import Preprocess
            self.preprocess = Preprocess(verbose=verbose)
        else:
            self.preprocess = preprocess

        self.log_interval = log_interval
        self.verbose = verbose
        if verbose:
            logger.set_level("DEBUG")
        else:
            logger.set_level("WARNING")

        logger.info(f'use device: {device}')

    def make_optimizer(self, learning_rate: float):
        args_dict = {
            'params': self.model.parameters(),
            'lr': learning_rate,
        }

        optimizer = torch.optim.Adam(**args_dict)
        return optimizer

    def fit(
        self,
        docs: List[str],
        epochs: int = 200,
        learning_rate: float = 0.002,
        preset_doc_embeddings: np.ndarray = None,
    ):
        self.fit_transform(docs, epochs, learning_rate, preset_doc_embeddings)
        return self

    def fit_transform(
        self,
        docs: List[str],
        epochs: int = 200,
        learning_rate: float = 0.002,
        preset_doc_embeddings: np.ndarray = None,
    ):
        """
        epochs: The number of epochs.
        learning_rate: The learning rate.
        low_memory: Set this to True to learn with a batch of documents at each time.
                     This uses less memory.
        batch_size: The batch size is used when low_memory is True.
        """
        # Preprocess docs
        data_size = len(docs)
        if self.low_memory:
            logger.info("Using low memory mode.")
            assert self.low_memory_batch_size is not None
            self.batch_size = self.low_memory_batch_size
            dataset_device = 'cpu'
        else:
            self.batch_size = data_size
            dataset_device = self.device

        # Fine-tune the model if it is already fitted.
        if check_fitted(self):
            logger.info("Fine-tuning the model.")
            _fitted = True
        else:
            logger.info("First fit the model.")
            _fitted = False

        # Create doc_embedder.
        self.doc_embedder = DocEmbedModel(self.doc_embed_model, self.device, self.normalize_embeddings, self.verbose)

        # Create the dataset.
        dataset = Dataset(
            docs,
            doc_embedder=self.doc_embedder,
            preprocess=self.preprocess,
            batch_size=self.batch_size,
            device=dataset_device,
            low_memory=self.low_memory,
            preset_doc_embeddings=preset_doc_embeddings
        )

        self.train_doc_embeddings = torch.as_tensor(dataset.doc_embeddings)
        if not self.low_memory:
            self.train_doc_embeddings = self.train_doc_embeddings.to(self.device)

        vocab_size = dataset.vocab_size
        doc_embed_size = dataset.doc_embed_size

        if not _fitted:
            self.model.init(vocab_size, doc_embed_size)
        else:
            pre_vocab = self.vocab
            self.model.init(
                vocab_size,
                doc_embed_size,
                _fitted,
                pre_vocab,
                dataset.vocab
            )

        self.vocab = dataset.vocab
        self.model = self.model.to(self.device)

        optimizer = self.make_optimizer(learning_rate)

        # Start training.
        self.model.train()
        for epoch in tqdm(range(1, epochs + 1), desc="Training FASTopic"):
            loss_rst_dict = defaultdict(float)

            for batch_bow, batch_doc_embed in dataset.dataloader:
                if self.low_memory:
                    batch_doc_embed = batch_doc_embed.to(self.device)
                    batch_bow = batch_bow.to(self.device)

                rst_dict = self.model(batch_bow, batch_doc_embed)
                batch_loss = rst_dict["loss"]

                optimizer.zero_grad()
                batch_loss.backward()
                optimizer.step()

                for key in rst_dict:
                    loss_rst_dict[key] += rst_dict[key] * batch_bow.shape[0]

            if epoch % self.log_interval == 0:
                output_log = f"Epoch: {epoch:03d}"
                for key in loss_rst_dict:
                    output_log += f" {key}: {loss_rst_dict[key] / data_size :.3f}"
                logger.info(output_log)

        self.beta = self.get_beta()
        self.top_words = self.get_top_words(self.num_top_words)
        self.train_theta = self.transform(self, self.train_doc_embeddings)

        return self.top_words, self.train_theta

    def transform(
            self,
            docs: List[str]=None,
            doc_embeddings: np.ndarray=None
        ):

        if docs is None and doc_embeddings is None:
            raise ValueError("Must set either docs or doc_embeddings.")

        if doc_embeddings is None and self.doc_embedder is None:
            raise ValueError("Must set doc embeddings.")

        if doc_embeddings is None:
            doc_embeddings = torch.as_tensor(self.doc_embedder.encode(docs))
            if not self.low_memory:
                doc_embeddings = doc_embeddings.to(self.device)

        with torch.no_grad():
            self.model.eval()
            theta = self.model.get_theta(doc_embeddings, self.train_doc_embeddings)
            theta = theta.detach().cpu().numpy()

        return theta

    def get_beta(self):
        """
            return beta: topic-word distributions matrix, $K \times V$
        """
        beta = self.model.get_beta().detach().cpu().numpy()
        return beta

    def get_top_words(self, num_top_words=15, verbose=None):
        if verbose is None:
            verbose = self.verbose
        beta = self.get_beta()
        top_words = get_top_words(beta, self.vocab, num_top_words, verbose)
        return top_words

    @property
    def topic_embeddings(self):
        """
            return topic embeddings $K \times L$
        """
        return self.model.topic_embeddings.detach().cpu().numpy()

    @property
    def word_embeddings(self):
        """
            return word embeddings $V \times L$
        """
        return self.model.word_embeddings.detach().cpu().numpy()

    @property
    def transp_DT(self):
        """
            return transp_DT $N \times K$
        """
        return self.model.get_transp_DT(self.train_doc_embeddings)

    def save(
        self,
        path: str
    ):
        """Saves the FASTopic model and its PyTorch model weights to the specified path, like `./fastopic.zip`.

        This method saves the dict attributes of the FASTopic object (`self`) except for `doc_embedder` for lower size.

        Args:
            path (str): The path to save the model files. If the directory doesn't exist, it will be created.

        Returns:
            None
        """
        assert_fitted(self)

        path = Path(path)
        parent_dir = path.parent
        if not parent_dir.exists():
            parent_dir.mkdir(parents=True, exist_ok=True)

        instance_dict = {}
        for key, value in self.__dict__.items():
            if key not in ['doc_embedder']:
                instance_dict[key] = value

        state = {
            "instance_dict": instance_dict
        }
        torch.save(state, path)

    @classmethod
    def from_pretrained(
            cls,
            path: str,
            preprocess: Callable = None,
            low_memory: bool = None,
            low_memory_batch_size: int = None,
            device: str=None
        ):
        """Loads a pre-trained FASTopic model from a saved file.

        This method loads a previously saved FASTopic model instance, and rebuilds the `doc_embedder`.

        Args:
            path: The path to the directory containing the serialized FASTopic object `fastopic.zip`.
            device: Move the loaded model to the device.
                    Make sure that the device is the same
                    if FASTopic was saved with your own document embedding model.
                    For instance, you must set `device="cuda"` if your own document embedding models was on GPU.
                    The device can be others if you use models in sentence-transformers.
        Returns:
            FASTopic: An instance of the FASTopic class loaded from the provided file.

        Raises:
            FileNotFoundError: If the specified `path` does not exist.
        """

        if device is None:
            device = "cuda" if torch.cuda.is_available() else "cpu"

        state = torch.load(path, map_location=device, weights_only=False)
        instance_dict = state["instance_dict"]
        instance_dict["device"] = device

        if preprocess:
            instance_dict["preprocess"] = preprocess
        if low_memory:
            instance_dict["low_memory"] = low_memory
            instance_dict["low_memory_batch_size"] = low_memory_batch_size

        instance = cls.__new__(cls)
        instance.__dict__.update(instance_dict)

        instance.doc_embedder = DocEmbedModel(
            instance_dict["doc_embed_model"],
            device=instance_dict["device"],
            normalize_embeddings=instance_dict["normalize_embeddings"]
        )

        if instance.verbose:
            logger.set_level("DEBUG")
        else:
            logger.set_level("WARNING")

        return instance

    def get_topic(
            self,
            topic_idx: int,
            num_top_words: int=5
        ):

        assert_fitted(self)
        words = self.top_words[topic_idx].split()[:num_top_words]
        scores = np.sort(self.beta[topic_idx])[:-(num_top_words + 1):-1]

        return tuple(zip(words, scores))

    def get_topic_weights(self):
        assert_fitted(self)
        topic_weights = self.transp_DT.sum(0)
        return topic_weights

    def visualize_topic(self, **args):
        assert_fitted(self)
        return _plot.visualize_topic(self, **args)

    def visualize_topic_hierarchy(self, **args):
        assert_fitted(self)
        return _plot.visualize_hierarchy(self, **args)

    def topic_activity_over_time(self,
                                 time_slices: List[int],
                                ):
        assert_fitted(self)
        topic_activity = self.transp_DT
        topic_activity *= self.transp_DT.shape[0]

        assert len(time_slices) == topic_activity.shape[0]

        df = pd.DataFrame(topic_activity)
        df['time_slices'] = time_slices
        topic_activity = df.groupby('time_slices').mean().to_numpy().transpose()

        return topic_activity

    def visualize_topic_activity(self, **args):
        assert_fitted(self)
        return _plot.visualize_activity(self, **args)

    def visualize_topic_weights(self, **args):
        assert_fitted(self)
        return _plot.visualize_topic_weights(self, **args)



import pytest
import shutil
from numpy.testing import assert_almost_equal
from topmost import Preprocess, download_dataset, DynamicDataset

import sys
sys.path.append('../')

from fastopic import FASTopic


@pytest.fixture
def cache_path():
    return './pytest_cache/'


@pytest.fixture
def num_topics():
    return 10


def test(cache_path, num_topics):
    download_dataset('NYT', cache_path=f'{cache_path}/datasets')
    dataset = DynamicDataset("./datasets/NYT", as_tensor=False)
    docs = dataset.train_texts

    model = FASTopic(num_topics, device='cpu')
    model.fit_transform(docs, epochs=1)
    beta = model.get_beta()

    path = f"{cache_path}/tmp_save/fastopic.zip"
    model.save(path)

    new_preprocess = Preprocess(vocab_size=200)
    new_model = FASTopic.from_pretrained(path, device='cuda', preprocess=new_preprocess)
    new_model.transform(dataset.test_texts)
    new_beta = new_model.get_beta()
    assert_almost_equal(beta, new_beta)

    new_model.fit_transform(docs, epochs=1)

    shutil.rmtree(f"{cache_path}/tmp_save")



import torch
from torch import nn
import torch.nn.functional as F
from ._model_utils import pairwise_euclidean_distance


class ETP(nn.Module):
    def __init__(
        self,
        sinkhorn_alpha,
        init_a_dist=None,
        init_b_dist=None,
        OT_max_iter=5000,
        stop_thr=0.5e-2,
    ):
        super().__init__()
        self.sinkhorn_alpha = sinkhorn_alpha
        self.OT_max_iter = OT_max_iter
        self.stop_thr = stop_thr
        self.init_a_dist = init_a_dist
        self.init_b_dist = init_b_dist

        if init_a_dist is not None:
            self.a_dist = init_a_dist

        if init_b_dist is not None:
            self.b_dist = init_b_dist

    def forward(self, x, y):
        # Sinkhorn's algorithm in log-space
        M = pairwise_euclidean_distance(x, y)
        device = M.device

        if self.init_a_dist is None:
            a = (torch.ones(M.shape[0]) / M.shape[0]).unsqueeze(1).to(device)
        else:
            a = F.softmax(self.a_dist, dim=0).to(device)

        if self.init_b_dist is None:
            b = (torch.ones(M.shape[1]) / M.shape[1]).unsqueeze(1).to(device)
        else:
            b = F.softmax(self.b_dist, dim=0).to(device)

        # LOG-DOMAIN STABILIZATION: Work in log-space to avoid extreme scaling values
        log_a = torch.log(a + 1e-30)  # Shape: (n, 1)
        log_b = torch.log(b + 1e-30)  # Shape: (m, 1)

        # Initialize log-domain scaling vectors
        log_u = torch.zeros_like(log_a)  # Shape: (n, 1)
        log_v = torch.zeros_like(log_b)  # Shape: (m, 1)

        log_K = -M * self.sinkhorn_alpha  # Shape: (n, m)

        # Sinkhorn iterations in log-domain
        err = 1
        cpt = 0
        while err > self.stop_thr and cpt < self.OT_max_iter:
            # LOG-DOMAIN UPDATE: log(v) = log(b) - logsumexp(log(K^T) + log(u), dim=0)
            # This is equivalent to: v = b / (K^T @ u) but numerically stable
            log_Ku = log_K.T + log_u.T  # Shape: (m, n)
            log_v = log_b - torch.logsumexp(log_Ku, dim=1).unsqueeze(1)  # Shape: (m, 1)

            # LOG-DOMAIN UPDATE: log(u) = log(a) - logsumexp(log(K) + log(v^T), dim=1)
            # This is equivalent to: u = a / (K @ v) but numerically stable
            log_Kv = log_K + log_v.T  # Shape: (n, m)
            log_u = log_a - torch.logsumexp(log_Kv, dim=1).unsqueeze(1)  # Shape: (n, 1)

            cpt += 1
            if cpt % 50 == 1:
                # Absorb current scalings
                log_K = log_K + log_u + log_v.T
                # Reset scalings
                log_u = torch.zeros_like(log_a)
                log_v = torch.zeros_like(log_b)

                err = self.check_convergence(log_K, log_u, log_v, a, b)

        # Convert final results back to linear domain for compatibility
        u = torch.exp(log_u)  # Shape: (n, 1)
        v = torch.exp(log_v)  # Shape: (m, 1)
        K = torch.exp(log_K)  # Shape: (n, m)

        # Compute transport plan: P = diag(u) K diag(v)
        transp = u * (K * v.T)  # Shape: (n, m)

        # Compute transport cost: <P, M>
        loss_ETP = torch.sum(transp * M)

        return loss_ETP, transp

    def check_convergence(self, log_K, log_u, log_v, a, b):
        """
        Check convergence by verifying the marginal constraints using absolute error.
        The transport plan is P = diag(u) @ K @ diag(v)
        We need: P @ 1 = a and P^T @ 1 = b
        """
        with torch.no_grad():
            # Row constraint: (u ⊙ (K @ v)) should equal a
            # In log domain: log(u) + log(K @ v) should equal log(a)
            log_Kv = log_K + log_v.T  # Shape: (n, m)
            log_Kv_sum = torch.logsumexp(
                log_Kv, dim=1, keepdim=True
            )  # log(K @ v), Shape: (n, 1)
            log_row_sums = log_u + log_Kv_sum  # log(u ⊙ (K @ v)), Shape: (n, 1)
            row_sums = torch.exp(log_row_sums)  # Shape: (n, 1)

            # Column constraint: (v ⊙ (K^T @ u)) should equal b
            # In log domain: log(v) + log(K^T @ u) should equal log(b)
            log_Ku = log_K.T + log_u.T  # Shape: (m, n)
            log_Ku_sum = torch.logsumexp(
                log_Ku, dim=1, keepdim=True
            )  # log(K^T @ u), Shape: (m, 1)
            log_col_sums = log_v + log_Ku_sum  # log(v ⊙ (K^T @ u)), Shape: (m, 1)
            col_sums = torch.exp(log_col_sums)  # Shape: (m, 1)

            # Compute absolute errors
            row_err = torch.abs(row_sums - a).sum()

            col_err = torch.abs(col_sums - b).sum()

            return max(row_err.item(), col_err.item())



import numpy as np
import torch
from sentence_transformers import SentenceTransformer

import logging
from typing import List, Callable


def get_top_words(beta, vocab, num_top_words, verbose=False):
    topic_str_list = list()
    for i, topic_dist in enumerate(beta):
        topic_words = np.array(vocab)[np.argsort(topic_dist)][:-(num_top_words + 1):-1]
        topic_str = ' '.join(topic_words)
        topic_str_list.append(topic_str)
        if verbose:
            print('Topic {}: {}'.format(i, topic_str))

    return topic_str_list


class DocEmbedModel:
    def __init__(
        self,
        model: str="all-MiniLM-L6-v2",
        device: str="cpu",
        normalize_embeddings: bool = False,
        verbose: bool = False,
    ):
        self.verbose = verbose
        self.normalize_embeddings = normalize_embeddings

        if isinstance(model, str):
            self.model = SentenceTransformer(model, device=device)
        else:
            self.model = model

    def encode(self, docs: List[str]) -> np.ndarray:
        embeddings = self.model.encode(
            docs,
            show_progress_bar=self.verbose,
            normalize_embeddings=self.normalize_embeddings
        )
        return embeddings


class DataIterator:
    def __init__(
        self,
        bow,
        doc_embeddings,
        batch_size,
        device,
        low_memory: bool = False,
        shuffle=True,
    ):
        """
        Args:
            bow: A sparse matrix (scipy.sparse matrix) representing bag-of-words data.
            doc_embeddings: A dense matrix (NumPy array or PyTorch tensor) representing document embeddings.
            batch_size: The size of each batch.
            device: The target device for loading data ('cpu' or 'cuda').
            shuffle: Whether to shuffle the data before batching (default: True).
        """
        self.doc_embeddings = torch.tensor(doc_embeddings, dtype=torch.float32)
        self.low_memory = low_memory
        self.shuffle = shuffle
        self.batch_size = batch_size

        if low_memory:
            self.bow = bow
        else:
            # If memory is enough, convert to tensors.
            self.bow = torch.tensor(bow.toarray(), dtype=torch.float32).to(device)
            self.doc_embeddings = self.doc_embeddings.to(device)

    def __iter__(self):
        """
        Define the iterator logic for batching data.

        Yields:
            A batch of sparse matrix tensors (batch_bow) and dense matrix tensors (batch_doc_embed).
        """
        num_samples = self.bow.shape[0]
        indices = np.arange(num_samples)

        if self.shuffle:
            np.random.shuffle(indices)

        # Generate batches.
        for start_idx in range(0, num_samples, self.batch_size):
            batch_indices = indices[start_idx:start_idx + self.batch_size]

            if self.low_memory:
                batch_bow = torch.tensor(self.bow[batch_indices].toarray(), dtype=torch.float32)
            else:
                batch_bow = self.bow[batch_indices]

            batch_doc_embed = self.doc_embeddings[batch_indices]
            
            yield batch_bow, batch_doc_embed


class Dataset:
    def __init__(
        self,
        docs: List[str],
        doc_embedder: Callable,
        preprocess: Callable,
        batch_size: int=200,
        device: str="cpu",
        low_memory: bool = False,
        preset_doc_embeddings=None
    ):
        rst = preprocess.preprocess(docs)
        self.train_bow = rst['train_bow']
        self.vocab = rst['vocab']

        self.vocab_size = len(self.vocab)

        if preset_doc_embeddings is None:
            self.doc_embeddings = doc_embedder.encode(docs)
        else:
            self.doc_embeddings = preset_doc_embeddings

        self.doc_embed_size = self.doc_embeddings.shape[1]
        self.dataloader = DataIterator(self.train_bow, self.doc_embeddings, batch_size, device, low_memory)


def check_fitted(model):
    """ Checks if the model was fitted by verifying the presence of self.beta.
    """
    return model.beta is not None


def assert_fitted(model):
    """ assert that the model was fitted.

    Arguments:
        model: FASTopic instance for which the check is performed.

    Returns:
        None

    Raises:
        ValueError: If the beta is not found.
    """
    msg = ("This %(name)s instance is not fitted yet. Call 'fit' with appropriate arguments before using this estimator.")

    if not check_fitted(model):
        raise ValueError(msg % {'name': type(model).__name__})


class Logger:
    def __init__(self, level):
        self.logger = logging.getLogger('FASTopic')
        self.set_level(level)
        self._add_handler()
        self.logger.propagate = False

    def info(self, message):
        self.logger.info(f"{message}")

    def warning(self, message):
        self.logger.warning(f"WARNING: {message}")

    def set_level(self, level):
        levels = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
        if level in levels:
            self.logger.setLevel(level)

    def _add_handler(self):
        sh = logging.StreamHandler()
        sh.setFormatter(logging.Formatter('%(asctime)s - %(name)s - %(message)s'))
        self.logger.addHandler(sh)

        # Remove duplicate handlers
        if len(self.logger.handlers) > 1:
            self.logger.handlers = [self.logger.handlers[0]]



import pytest
import shutil
from numpy.testing import assert_almost_equal
from topmost import download_dataset, DynamicDataset
from sentence_transformers import SentenceTransformer
from typing import List

import sys
sys.path.append('../')

from fastopic import FASTopic


@pytest.fixture
def cache_path():
    return './pytest_cache/'


@pytest.fixture
def num_topics():
    return 10


class DocEmbedModel:
    def __init__(
            self,
            model: str="all-MiniLM-L6-v2",
            device: str='cpu'
        ):

        self.doc_embed_model = SentenceTransformer(model, device=device)

    def encode(self,
               docs:List[str],
               show_progress_bar:bool=False,
               normalize_embeddings: bool=True
            ):

        embeddings = self.doc_embed_model.encode(
            docs,
            show_progress_bar=show_progress_bar,
            normalize_embeddings=normalize_embeddings,
        )
        return embeddings


def test(cache_path, num_topics):
    download_dataset("NYT", cache_path=f"{cache_path}/datasets")
    dataset = DynamicDataset("./datasets/NYT", as_tensor=False)

    custom_model = DocEmbedModel()

    model = FASTopic(num_topics, doc_embed_model=custom_model)
    docs = dataset.train_texts
    model.fit_transform(docs, epochs=1)
    beta = model.get_beta()

    path = f"{cache_path}/tmp_save/fastopic.zip"
    model.save(path)

    new_model = FASTopic.from_pretrained(path)
    new_beta = new_model.get_beta()
    assert_almost_equal(beta, new_beta)

    new_model.fit_transform(docs, epochs=1)

    shutil.rmtree(f"{cache_path}/tmp_save")
